﻿using System;
using System.ComponentModel;
using System.Windows.Forms;
using System.IO;

using NAudio.Wave;

namespace msxdisk
{
    public partial class PlayCASForm : Form
    {
        class WaveProvider : WaveProvider32
        {
            PlayCASForm form;

            public WaveProvider(PlayCASForm form)
            {
                this.SetWaveFormat(9600, 1);
                this.form = form;
            }

            public override int Read(float[] buffer, int offset, int sampleCount)
            {
                return form.Read(buffer, offset, sampleCount);
            }
        }

        byte[] header = new byte[] { 0x1f, 0xa6, 0xde, 0xba, 0xcc, 0x13, 0x7d, 0x74 };

        float[] samplesX = new float[] { 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f };
        float[] samples0 = new float[] { 1f, 1f, 1f, 1f, -1f, -1f, -1f, -1f };
        float[] samples1 = new float[] { 1f, 1f, -1f, -1f, 1f, 1f, -1f, -1f };


        public string casfile;
        FileStream fsCAS;
        byte[] casBuffer = new byte[512];
        int casBufferBytes;
        int casBufferPos;

        int cSilence;
        int cHeader;

        WaveOut waveOut;
        WaveProvider waveProvider;

        bool play = false;

        public PlayCASForm()
        {
            InitializeComponent();
        }

        private void PlayCASForm_Load(object sender, EventArgs e)
        {
            this.Text = "Playing " + Path.GetFileName(casfile) + "...";

            try
            {
                fsCAS = new FileStream(casfile, FileMode.Open, FileAccess.Read);

                init();

                waveProvider = new WaveProvider(this);

                waveOut = new WaveOut();
                waveOut.DesiredLatency = 100;
                waveOut.Init(waveProvider);
                waveOut.Play();

                Success();
            }
            catch (Exception ex)
            {
                Error(ex.Message);
            }
        }

        private void PlayCASForm_Shown(object sender, EventArgs e)
        {

        }

        private void PlayCASForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            timer.Stop();
            if (waveOut != null)
                waveOut.Stop();
            if (fsCAS != null)
                fsCAS.Close();
        }

        private void btnPlay_Click(object sender, EventArgs e)
        {
            play = true;
        }

        private void btnStop_Click(object sender, EventArgs e)
        {
            play = false;
        }

        private void btnRewind_Click(object sender, EventArgs e)
        {
            play = false;
            init();
        }
        
        void init()
        {
            lock(this)
            {
                fsCAS.Seek(0, SeekOrigin.Begin);
                casBufferBytes = casBuffer.Length;
                casBufferPos = casBuffer.Length;
                cSilence = 0;
                cHeader = 0;
            }
        }

        int pos()
        {
            lock (this)
            {
                return (int)fsCAS.Position - casBufferBytes + casBufferPos;
            }
        }

        bool is_header()
        {
            if (casBufferPos + 8 > casBufferBytes)
                return false;
            for (int i = 0; i < 8; i++)
                if (casBuffer[casBufferPos + i] != header[i])
                    return false;
            return true;
        }

        int writeX(float[] buffer, ref int offset)
        {
            samplesX.CopyTo(buffer, offset);
            offset += samplesX.Length;
            return samplesX.Length;
        }

        int write0(float[] buffer, ref int offset)
        {
            samples0.CopyTo(buffer, offset);
            offset += samples0.Length;
            return samples0.Length;
        }

        int write1(float[] buffer, ref int offset)
        {
            samples1.CopyTo(buffer, offset);
            offset += samples1.Length;
            return samples1.Length;
        }

        float[] q = new float[960];
        int qLen = 0;
        int i = 0;

        int Read(float[] buffer, int offset, int sampleCount)
        {
            for (int j = 0; ; i = 0)
            {
                for (; i < qLen; )
                {
                    buffer[offset + j] = q[i];
                    i++;
                    j++;
                    if (j == sampleCount)
                        return sampleCount;
                }

                for(qLen = 0; q.Length - qLen >= 88; )
                {
                    if (play)
                    {
                        lock (this)

                            if (cSilence > 0) //silence
                            {
                                writeX(q, ref qLen);
                                cSilence--;
                            }
                            else if (cHeader > 0) //header
                            {
                                write1(q, ref qLen);
                                cHeader--;
                                if (cHeader == 0) casBufferPos += 8;
                            }
                            else if (casBufferPos == casBuffer.Length) //end of q
                            {
                                casBufferBytes = fsCAS.Read(casBuffer, 0, casBuffer.Length);
                                casBufferPos = 0;
                            }
                            else if (casBufferPos == casBufferBytes) //not end of q AND end of q bytes, therefore eof
                            {
                                play = false; //stop playback
                            }
                            else if (casBufferPos % 8 == 0 && is_header()) //header
                            {
                                if (pos() != 0) cSilence = 1200;
                                cHeader = 2000;
                            }
                            else // data
                            {
                                int data = casBuffer[casBufferPos];
                                write0(q, ref qLen);
                                for (int i = 0; i < 8; i++)
                                {
                                    if ((data & 0x01) == 0)
                                        write0(q, ref qLen);
                                    else
                                        write1(q, ref qLen);
                                    data = data >> 1;
                                }
                                write1(q, ref qLen);
                                write1(q, ref qLen);

                                casBufferPos++;
                            }
                    }
                    else
                    {
                        writeX(q, ref qLen);
                    }
                }
            }
        }


        private void timer_Tick(object sender, EventArgs e)
        {
            if (fsCAS != null)
            {
                SetProgress((int)((float)pos() / (float)fsCAS.Length * 1000F));
            }
        }

        delegate void SetProgressCallback(int progress);

        private void SetProgress(int progress)
        {
            if (this.progress.InvokeRequired)
            {
                SetProgressCallback d = new SetProgressCallback(SetProgress);
                this.Invoke(d, new object[] { progress });
            }
            else
            {
                this.progress.Value = progress;
            }
        }

        delegate void ErrorCallback(string message);

        private void Error(string message)
        {
            if (this.InvokeRequired)
            {
                ErrorCallback d = new ErrorCallback(Error);
                this.Invoke(d, new object[] { message });
            }
            else
            {
                MessageBox.Show(this, message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                this.Close();
            }
        }

        delegate void SuccessCallback();

        private void Success()
        {
            if (this.InvokeRequired)
            {
                SuccessCallback d = new SuccessCallback(Success);
                this.Invoke(d, new object[] {});
            }
            else
            {
                timer.Start();
                pnlControls.Enabled = true;
                this.Cursor = Cursors.Default;
            }
        }
    }
}
